@@ -84,7 +84,7 @@ NULL
 #' @importFrom igraph contract delete_edges delete_vertex_attr which_loop
 #' which_multiple
 #' @importFrom sf st_as_sf st_cast st_centroid st_combine st_geometry
-#' st_intersects
+#' st_geometry<- st_intersects
 #' @importFrom tibble as_tibble
 #' @importFrom tidygraph as_tbl_graph
 #' @export
@@ -161,7 +161,7 @@ to_spatial_contracted = function(x, ..., simplify = FALSE,
   }
   # Update the nodes table of the contracted network.
   new_nodes = st_as_sf(new_nodes, sf_column_name = geom_colname)
-  node_graph_attributes(x_new) = new_nodes
+  node_attribute_values(x_new) = new_nodes
   # Convert in a sfnetwork.
   x_new = tbg_to_sfn(x_new)
   ## ===============================================================
@@ -179,7 +179,7 @@ to_spatial_contracted = function(x, ..., simplify = FALSE,
     x_new = delete_edges(x_new, which(which_loop(x_new)))
     x_new = x_new %preserve_all_attrs% x_new
   }
-  if (has_spatially_explicit_edges(x)) {
+  if (has_explicit_edges(x)) {
     # Extract the edges and their geometries from the contracted network.
     new_edges = edges_as_sf(x_new)
     new_edge_geoms = st_geometry(new_edges)
@@ -291,11 +291,11 @@ to_spatial_contracted = function(x, ..., simplify = FALSE,
     new_edge_geoms[E3] = geoms
     # Update the edges table of the contracted network.
     st_geometry(new_edges) = new_edge_geoms
-    edge_graph_attributes(x_new) = new_edges
+    edge_attribute_values(x_new) = new_edges
   }
   # Return in a list.
   list(
-    contracted = x_new %preserve_graph_attrs% x
+    contracted = x_new %preserve_network_attrs% x
   )
 }
 
@@ -310,7 +310,7 @@ to_spatial_contracted = function(x, ..., simplify = FALSE,
 #' @importFrom igraph is_directed
 #' @export
 to_spatial_directed = function(x) {
-  require_spatially_explicit_edges(x)
+  require_explicit_edges(x)
   if (is_directed(x)) return (x)
   # Retrieve the nodes and edges from the network.
   nodes = nodes_as_sf(x)
@@ -328,7 +328,7 @@ to_spatial_directed = function(x) {
   x_new = sfnetwork_(nodes, edges, directed = TRUE)
   # Return in a list.
   list(
-    directed = x_new %preserve_graph_attrs% x
+    directed = x_new %preserve_network_attrs% x
   )
 }
 
@@ -351,7 +351,7 @@ to_spatial_explicit = function(x, ...) {
     edges = edges_as_table(x)
     new_edges = st_as_sf(edges, ...)
     x_new = x
-    edge_graph_attributes(x_new) = new_edges
+    edge_attribute_values(x_new) = new_edges
   } else {
     x_new = explicitize_edges(x)
   }
@@ -398,10 +398,9 @@ to_spatial_explicit = function(x, ...) {
 to_spatial_neighborhood = function(x, node, threshold, weights = NULL,
                                    from = TRUE, ...) {
   # Parse node argument.
-  # If 'node' is given as simple feature geometry, convert it to a node index.
-  # This can be done equal to setting endpoints of path calculations.
+  # If 'node' is given as a geometry, find the index of the nearest node.
   # When multiple nodes are given only the first one is taken.
-  if (is.sf(node) | is.sfc(node)) node = set_path_endpoints(x, node)
+  if (is.sf(node) | is.sfc(node)) node = get_nearest_node_index(x, node)
   if (length(node) > 1) raise_multiple_elements("node")
   # Parse weights argument.
   # This can be done equal to setting weights for path calculations.
@@ -468,7 +467,7 @@ to_spatial_shortest_paths = function(x, ...) {
 #' @param remove_loops Should loop edges be removed. Defaults to \code{TRUE}.
 #'
 #' @importFrom igraph simplify
-#' @importFrom sf st_as_sf st_crs st_sfc
+#' @importFrom sf st_as_sf st_crs st_crs<- st_precision st_precision<- st_sfc
 #' @importFrom tibble as_tibble
 #' @importFrom tidygraph as_tbl_graph
 #' @export
@@ -477,7 +476,7 @@ to_spatial_simple = function(x, remove_multiple = TRUE, remove_loops = TRUE,
                              store_original_data = FALSE) {
   # Define if the network has spatially explicit edges.
   # This influences some of the processes to come.
-  spatial = if (has_spatially_explicit_edges(x)) TRUE else FALSE
+  spatial = if (has_explicit_edges(x)) TRUE else FALSE
   # Update the attribute summary instructions.
   # In the summarise attributes only real attribute columns were referenced.
   # On top of those, we need to include:
@@ -499,7 +498,7 @@ to_spatial_simple = function(x, remove_multiple = TRUE, remove_loops = TRUE,
     remove.multiple = remove_multiple,
     remove.loops = remove_loops,
     edge.attr.comb = summarise_attributes
-  ) %preserve_graph_attrs% x
+  ) %preserve_network_attrs% x
   # Igraph does not know about geometry list columns.
   # Summarizing them results in a list of sfg objects.
   # We should reconstruct the sfc geometry list column out of that.
@@ -508,7 +507,8 @@ to_spatial_simple = function(x, remove_multiple = TRUE, remove_loops = TRUE,
     new_edges[geom_colname] = list(st_sfc(new_edges[[geom_colname]]))
     new_edges = st_as_sf(new_edges, sf_column_name = geom_colname)
     st_crs(new_edges) = st_crs(x)
-    edge_graph_attributes(x_new) = new_edges
+    st_precision(new_edges) = st_precision(x)
+    edge_attribute_values(x_new) = new_edges
   }
   # If requested, original edge data should be stored in a .orig_data column.
   if (store_original_data) {
@@ -517,7 +517,7 @@ to_spatial_simple = function(x, remove_multiple = TRUE, remove_loops = TRUE,
     new_edges = edges_as_table(x_new)
     copy_data = function(i) edges[i, , drop = FALSE]
     new_edges$.orig_data = lapply(new_edges$.tidygraph_edge_index, copy_data)
-    edge_graph_attributes(x_new) = new_edges
+    edge_attribute_values(x_new) = new_edges
   }
   # Return in a list.
   list(
@@ -529,24 +529,56 @@ to_spatial_simple = function(x, remove_multiple = TRUE, remove_loops = TRUE,
 #' iteratively removing pseudo nodes, while preserving the connectivity of the
 #' network. In the case of directed networks, pseudo nodes are those nodes that
 #' have only one incoming and one outgoing edge. In undirected networks, pseudo
-#' nodes are those nodes that have two incident edges. Connectivity of the
+#' nodes are those nodes that have two incident edges. Equality of attribute
+#' values among the two edges can be defined as an additional requirement by
+#' setting the \code{require_equal} parameter. Connectivity of the
 #' network is preserved by concatenating the incident edges of each removed
 #' pseudo node. Returns a \code{morphed_sfnetwork} containing a single element
 #' of class \code{\link{sfnetwork}}.
-#' @importFrom dplyr bind_rows
+#'
+#' @param protect Nodes to be protected from being removed, no matter if they
+#' are a pseudo node or not. Can be given as a numeric vector containing node
+#' indices or a character vector containing node names. Can also be a set of
+#' geospatial features as object of class code{\link[sf]{sf}} or
+#' \code{\link[sf]{sfc}}. In that case, for each of these features its nearest
+#' node in the network will be protected. Defaults to \code{NULL}, meaning that
+#' none of the nodes is protected.
+#'
+#' @param require_equal Should nodes only be removed when the attribute values
+#' of their incident edges are equal? Defaults to \code{FALSE}. If \code{TRUE},
+#' only pseudo nodes that have incident edges with equal attribute values are
+#' removed. May also be given as a vector of attribute names. In that case only
+#' those attributes are checked for equality. Equality tests are evaluated
+#' using the \code{==} operator.
+#'
 #' @importFrom igraph adjacent_vertices decompose degree delete_vertices
-#' edge_attr get.edge.ids induced_subgraph is_directed vertex_attr
-#' @importFrom sf st_as_sf st_cast st_combine st_crs st_equals st_line_merge
+#' edge_attr edge.attributes get.edge.ids igraph_opt igraph_options
+#' incident_edges induced_subgraph is_directed vertex_attr
+#' @importFrom sf st_as_sf st_cast st_combine st_crs st_equals st_is
+#' st_line_merge
 #' @export
-to_spatial_smooth = function(x, store_original_data = FALSE) {
+to_spatial_smooth = function(x,
+                             protect = NULL,
+                             summarise_attributes = "ignore",
+                             require_equal = FALSE,
+                             store_original_data = FALSE) {
+  # Change default igraph options.
+  # This prevents igraph returns node or edge indices as formatted sequences.
+  # We only need the "raw" integer indices.
+  # Changing this option can lead to quiet a performance improvement.
+  default_igraph_opt = igraph_opt("return.vs.es")
+  igraph_options(return.vs.es = FALSE)
+  on.exit(igraph_options(return.vs.es = default_igraph_opt))
   # Retrieve nodes and edges from the network.
   nodes = nodes_as_sf(x)
   edges = edges_as_table(x)
-  # Check whether:
-  # --> x is directed.
-  # --> x has spatially explicit edges.
+  # For later use:
+  # --> Check if x is directed.
+  # --> Check if x has spatially explicit edges.
+  # --> Retrieve the name of the geometry column of the edges in x.
   directed = is_directed(x)
   spatial = is.sf(edges)
+  geom_colname = attr(edges, "sf_column")
   ## ==========================
   # STEP I: DETECT PSEUDO NODES
   # The first step is to detect which nodes in x are pseudo nodes.
@@ -561,37 +593,133 @@ to_spatial_smooth = function(x, store_original_data = FALSE) {
     pseudo = degree(x) == 2
   }
   if (! any(pseudo)) return (x)
-  ## ===============================
-  # STEP II: FIND EDGES TO BE MERGED
-  # The connectivity of the network should be preserved.
-  # Therefore we need to:
-  # --> Find adjacent nodes of a pseudo node.
+  ## ===========================
+  # STEP II: FILTER PSEUDO NODES
+  # Users can define additional requirements for a node to be smoothed:
+  # --> It should not be listed in the provided set of protected nodes.
+  # --> Its incident edges should have equal values for some attributes.
+  # In these cases we need to filter the set of detected pseudo nodes.
+  ## ===========================
+  # Detected pseudo nodes that are protected should be filtered out.
+  if (! is.null(protect)) {
+    # Parse the protect parameter values.
+    # If protect is given as character vector:
+    # --> Find the node indices belonging to these node names.
+    # If protect is given as geospatial features:
+    # --> First find the nearest node to each of these features.
+    if (is.character(protect)) {
+      # Obtain node names.
+      # They should be stored in a node attribute column named "name".
+      node_names = vertex_attr(x, "name")
+      if (is.null(node_names)) {
+        stop(
+          "Node names should be stored in an attribute column called ",
+          sQuote("name"),
+          call. = FALSE
+        )
+      }
+      # Match node names to node indices.
+      matched_names = match(protect, node_names)
+      if (any(is.na(matched_names))) {
+        stop(
+          "Unknown node names: ",
+          paste(sQuote(protect[is.na(matched_names)]), collapse = " and "),
+          ". Make sure node names are stored in an attribute column called ",
+          sQuote("name"),
+          call. = FALSE
+        )
+      }
+      protect = matched_names
+    } else if (is.sf(protect) | is.sfc(protect)) {
+      protect = get_nearest_node_index(x, protect)
+    }
+    # Mark all protected nodes as not being a pseudo node.
+    pseudo[protect] = FALSE
+    if (! any(pseudo)) return (x)
+  }
+  # Check for equality of certain attributes between incident edges.
+  # Detected pseudo nodes that fail this check should be filtered out.
+  if (! isFALSE(require_equal)) {
+    # If require_equal is TRUE all attributes will be checked for equality.
+    # In other cases only a subset of attributes will be checked.
+    if (isTRUE(require_equal)) {
+      require_equal = edge_attribute_names(x)
+    } else {
+      # Check if all given attributes exist in the edges table of x.
+      attr_exists = require_equal %in% edge_attribute_names(x)
+      if (! all(attr_exists)) {
+        stop(
+          "Unknown edge attributes: ",
+          paste(sQuote(require_equal[!attr_exists]), collapse = " and "),
+          call. = FALSE
+        )
+      }
+    }
+    # Get the node indices of the detected pseudo nodes.
+    pseudo_idxs = which(pseudo)
+    # Get the edge indices of the incident edges of each pseudo node.
+    # Combine them into a single numerical vector.
+    # Note the + 1 since incident_edges returns indices starting from 0.
+    incident_idxs = incident_edges(x, pseudo_idxs, mode = "all")
+    incident_idxs = do.call("c", incident_idxs) + 1
+    # Define for each of the incident edges if they are incoming or outgoing.
+    # In undirected networks this can be read instead as "first or second".
+    is_in = seq(1, 2 * length(pseudo_idxs), by = 2)
+    is_out = seq(2, 2 * length(pseudo_idxs), by = 2)
+    # Obtain the attributes to be checked for each of the incident edges.
+    incident_attrs = edge.attributes(x, incident_idxs)[require_equal]
+    # For each of these attributes:
+    # --> Check if its value is equal for both incident edges of a pseudo node.
+    check_equality = function(A) {
+      # Check equality for each pseudo node.
+      # NOTE:
+      # --> Operator == is used because element-wise comparisons are needed.
+      # --> Not sure if this approach works with identical() or all.equal().
+      are_equal = A[is_in] == A[is_out]
+      # If one of the two values is NA or NaN:
+      # --> The result of the element-wise comparison is always NA.
+      # --> This means the two elements are certainly not equal.
+      # --> Hence the result of this comparison can be set to FALSE.
+      are_equal[is.na(are_equal)] = FALSE
+      are_equal
+    }
+    tests = lapply(incident_attrs, check_equality)
+    # If one or more equality tests failed for a detected pseudo node:
+    # --> Mark this pseudo node as FALSE, i.e. not being a pseudo node.
+    failed = rowSums(do.call("cbind", tests)) != length(require_equal)
+    pseudo[pseudo_idxs[failed]] = FALSE
+    if (! any(pseudo)) return (x)
+  }
+  ## ====================================
+  # STEP II: INITIALIZE REPLACEMENT EDGES
+  # When removing pseudo nodes their incident edges get removed to.
+  # To preserve the network connectivity we need to:
+  # --> Find the two adjacent nodes of a pseudo node.
   # --> Connect these by merging the incident edges of the pseudo node.
-  # However, an adjacent node can also be another pseudo node.
-  # Then, we need to look further until we find a non-pseudo (junction) node.
-  # Hence, instead of processing each pseudo node on its own, we need to:
+  # An adjacent node of a pseudo node can also be another pseudo node.
+  # Instead of processing each pseudo node on its own, we will:
   # --> Find connected sets of pseudo nodes.
-  # --> Find the adjacent junction nodes to that set.
-  # --> Connect these by merging the edges in the set *and* its incident edges.
-  ## ===============================
+  # --> Find the adjacent non-pseudo nodes (junction or pendant) to that set.
+  # --> Connect them by merging the edges in the set plus its incident edges.
+  ## ====================================
   # Subset x to only contain pseudo nodes and the edges between them.
   # Decompose this subgraph to find connected sets of pseudo nodes.
-  x_pseudo = decompose(induced_subgraph(x, pseudo))
+  pseudo_sets = decompose(induced_subgraph(x, pseudo))
   # For each set of connected pseudo nodes:
-  # --> Find the indices of the adjacent junction node(s).
+  # --> Find the indices of the adjacent nodes.
   # --> Find the indices of the edges that need to be merged.
-  find_edges = function(G) {
-    # Retrieve the original node indices of the pseudo nodes in this set.
-    # Retrieve the original edge indices of the edges that connect them.
-    N = vertex_attr(G, ".tidygraph_node_index")
-    E = edge_attr(G, ".tidygraph_edge_index")
-    # Find all required node and edge indices.
-    if (directed) {
+  # The workflow for this is different for directed and undirected networks.
+  if (directed) {
+    initialize_replacement_edge = function(S) {
+      # Retrieve the original node indices of the pseudo nodes in this set.
+      # Retrieve the original edge indices of the edges that connect them.
+      N = vertex_attr(S, ".tidygraph_node_index")
+      E = edge_attr(S, ".tidygraph_edge_index")
       # Find the following:
       # --> The index of the pseudo node where an edge comes into the set.
       # --> The index of the pseudo node where an edge goes out of the set.
-      n_i = N[degree(G, mode = "in") == 0]
-      n_o = N[degree(G, mode = "out") == 0]
+      n_i = N[degree(S, mode = "in") == 0]
+      n_o = N[degree(S, mode = "out") == 0]
       # If these nodes do not exists:
       # --> We are dealing with a loop of connected pseudo nodes.
       # --> The loop is by definition not connected to the rest of the network.
@@ -600,151 +728,214 @@ to_spatial_smooth = function(x, store_original_data = FALSE) {
       if (length(n_i) == 0) return (NULL)
       # Find the following:
       # --> The index of the edge that comes in to the pseudo node set.
-      # --> The index of the node at the other end of that edge.
+      # --> The index of the non-pseudo node at the other end of that edge.
       # We'll call this the source node and source edge of the set.
-      src_node = as.integer(adjacent_vertices(x, n_i, mode = "in"))
-      src_edge = get.edge.ids(x, c(src_node, n_i))
+      # Note the + 1 since adjacent_vertices returns indices starting from 0.
+      source_node = adjacent_vertices(x, n_i, mode = "in")[[1]] + 1
+      source_edge = get.edge.ids(x, c(source_node, n_i))
       # Find the following:
       # --> The index of the edge that goes out of the pseudo node set.
-      # --> The index of the node at the other end of that edge.
-      # We'll call this the target node and target edge of the set.
-      trg_node = as.integer(adjacent_vertices(x, n_o, mode = "out"))
-      trg_edge = get.edge.ids(x, c(n_o, trg_node))
-    } else {
-      # In an undirected network there is no in or out. Instead, we find:
-      # --> As source: The node with the lowest index connected to the set.
-      # --> As target: The node with the highest index connected to the set.
+      # --> The index of the non-pseudo node at the other end of that edge.
+      # We'll call this the sink node and sink edge of the set.
+      # Note the + 1 since adjacent_vertices returns indices starting from 0.
+      sink_node = adjacent_vertices(x, n_o, mode = "out")[[1]] + 1
+      sink_edge = get.edge.ids(x, c(n_o, sink_node))
+      # List indices of all edges that will be merged into the replacement edge.
+      edge_idxs = c(source_edge, E, sink_edge)
+      # Return all retrieved information in a list.
+      list(
+        from = as.integer(source_node),
+        to = as.integer(sink_node),
+        .tidygraph_edge_index = as.integer(edge_idxs)
+      )
+    }
+  } else {
+    initialize_replacement_edge = function(S) {
+      # Retrieve the original node indices of the pseudo nodes in this set.
+      # Retrieve the original edge indices of the edges that connect them.
+      N = vertex_attr(S, ".tidygraph_node_index")
+      E = edge_attr(S, ".tidygraph_edge_index")
+      # Find the following:
+      # --> The two adjacent non-pseudo nodes to the set.
+      # --> The edges that connect these nodes to the set.
+      # We'll call these the adjacent nodes and incident edges of the set.
+      # --> The adjacent node with the lowest index will be the source node.
+      # --> The adjacent node with the higest index will be the sink node.
       if (length(N) == 1) {
         # When we have a single pseudo node that forms a set:
-        # --> It will be adjacent to the source and target.
-        con_nodes = as.integer(adjacent_vertices(x, N)[[1]])
-        src_node = min(con_nodes)
-        src_edge = get.edge.ids(x, c(src_node, N))
-        trg_node = max(con_nodes)
-        trg_edge = get.edge.ids(x, c(N, trg_node))
+        # --> It will be adjacent to both adjacent nodes of the set.
+        # Note the + 1 since adjacent_vertices returns indices starting from 0.
+        adjacent = adjacent_vertices(x, N)[[1]] + 1
+        if (length(adjacent) == 1) {
+          # If there is only one adjacent node to the pseudo node:
+          # --> The two adjacent nodes of the set are the same node.
+          # --> We only have to query for incident edges of the set once.
+          incident = get.edge.ids(x, c(adjacent, N))
+          source_node = adjacent
+          source_edge = incident[1]
+          sink_node = adjacent
+          sink_edge = incident[2]
+        } else {
+          # If there are two adjacent nodes to the pseudo node:
+          # --> The one with the lowest index will be source node.
+          # --> The one with the highest index will be sink node.
+          source_node = min(adjacent)
+          source_edge = get.edge.ids(x, c(source_node, N))
+          sink_node = max(adjacent)
+          sink_edge = get.edge.ids(x, c(N, sink_node))
+        }
       } else {
         # When we have a set of multiple pseudo nodes:
         # --> There are two pseudo nodes that form the boundary of the set.
         # --> These are the ones connected to only one other pseudo node.
-        N_b = N[degree(G) == 1]
+        N_b = N[degree(S) == 1]
         # If these boundaries do not exist:
         # --> We are dealing with a loop of connected pseudo nodes.
         # --> The loop is by definition not connected to the rest of the network.
         # --> Hence, there is no need to create a new edge.
         # --> Therefore we should not return a path.
         if (length(N_b) == 0) return (NULL)
-        # Find the source/target node connected to the first set boundary.
-        # --> Its adjacent nodes will be one pseudo node and a source/target.
-        # --> The source/target node is the one not present in the pseudo set.
-        n_1 = N_b[1]
-        adj_nodes = as.integer(adjacent_vertices(x, n_1)[[1]])
-        con_node_1 = adj_nodes[!(adj_nodes %in% N)]
-        # Find the source/target node connected to the second set boundary.
-        # --> Its adjacent nodes will be a pseudo node and a source/target.
-        # --> The source/target node is the one not present in the pseudo set.
-        n_2 = N_b[2]
-        adj_nodes = as.integer(adjacent_vertices(x, n_2)[[1]])
-        con_node_2 = adj_nodes[!(adj_nodes %in% N)]
-        # Define which of found nodes is the source and which the target.
-        if (con_node_1 < con_node_2) {
-          src_node = con_node_1
-          src_edge = get.edge.ids(x, c(src_node, n_1))
-          trg_node = con_node_2
-          trg_edge = get.edge.ids(x, c(trg_node, n_2))
-        } else {
-          src_node = con_node_2
-          src_edge = get.edge.ids(x, c(src_node, n_2))
-          trg_node = con_node_1
-          trg_edge = get.edge.ids(x, c(trg_node, n_1))
+        # Find the adjacent nodes of the set.
+        # These are the adjacent non-pseudo nodes to the boundaries of the set.
+        # We find them iteratively for the two boundary nodes of the set:
+        # --> A boundary connects to one pseudo node and one non-pseudo node.
+        # --> The non-pseudo node is the one not present in the pseudo set.
+        # Note the + 1 since adjacent_vertices returns indices starting from 0.
+        get_set_neighbour = function(n) {
+          all = adjacent_vertices(x, n)[[1]] + 1
+          all[!(all %in% N)]
         }
+        adjacent = do.call("c", lapply(N_b, get_set_neighbour))
+        # The adjacent node with the lowest index will be source node.
+        # The adjacent node with the highest index will be sink node.
+        N_b = N_b[order(adjacent)]
+        source_node = min(adjacent)
+        source_edge = get.edge.ids(x, c(source_node, N_b[1]))
+        sink_node = max(adjacent)
+        sink_edge = get.edge.ids(x, c(N_b[2], sink_node))
       }
+      # List indices of all edges that will be merged into the replacement edge.
+      edge_idxs = c(source_edge, E, sink_edge)
+      # Return all retrieved information in a list.
+      list(
+        from = as.integer(source_node),
+        to = as.integer(sink_node),
+        .tidygraph_edge_index = as.integer(edge_idxs)
+      )
     }
-    # List all edge indices in the path.
-    edge_idxs = c(src_edge, E, trg_edge)
-    # Return all retrieved information in a list.
-    list(from = src_node, to = trg_node, .tidygraph_edge_index = edge_idxs)
   }
-  new_edge_list = lapply(x_pseudo, find_edges)
-  new_edge_list = new_edge_list[lengths(new_edge_list) != 0] # Remove NULLs.
-  # Create a data frame with the merged edges.
-  new_edges = data.frame(do.call("rbind", new_edge_list))
-  new_edges$from = as.integer(new_edges$from)
-  new_edges$to = as.integer(new_edges$to)
-  ## ====================================
-  # STEP III: CONCATENATE EDGE GEOMETRIES
-  # If the edges to be merged have geometries:
+  new_idxs = lapply(pseudo_sets, initialize_replacement_edge)
+  new_idxs = new_idxs[lengths(new_idxs) != 0] # Remove NULLs.
+  ## ===================================
+  # STEP III: SUMMARISE EDGE ATTRIBUTES
+  # Each replacement edge replaces multiple original edges.
+  # Their attributes should all be summarised in a single value.
+  # The summary techniques to be used are given as summarise_attributes.
+  ## ===================================
+  # Obtain the attribute values of all original edges in the network.
+  # These should not include the geometries and original edge indices.
+  exclude = c(".tidygraph_edge_index", geom_colname)
+  edge_attrs = edge.attributes(x)
+  edge_attrs = edge_attrs[!(names(edge_attrs) %in% exclude)]
+  # For each replacement edge:
+  # --> Summarise the attributes of the edges it replaces into single values.
+  merge_attrs = function(E) {
+    orig_edges = E$.tidygraph_edge_index
+    orig_attrs = lapply(edge_attrs, `[`, orig_edges)
+    apply_summary_function = function(i) {
+      # Store return value in a list.
+      # This prevents automatic type promotion when rowbinding later on.
+      list(get_summary_function(i, summarise_attributes)(orig_attrs[[i]]))
+    }
+    new_attrs = lapply(names(orig_attrs), apply_summary_function)
+    names(new_attrs) = names(orig_attrs)
+    new_attrs
+  }
+  new_attrs = lapply(new_idxs, merge_attrs)
+  ## ===================================
+  # STEP VI: CONCATENATE EDGE GEOMETRIES
+  # If the edges to be replaced have geometries:
   # --> These geometries have to be concatenated into a single new geometry.
-  # --> The new geometry should go from the defined source to target node.
-  ## ====================================
+  # --> The new geometry should go from the defined source to sink node.
+  ## ===================================
   if (spatial) {
-    # For each new edge:
-    # --> Merge all original edge geometries into a single geometry.
+    # Obtain geometries of all original edges and nodes in the network.
     edge_geoms = st_geometry(edges)
     node_geoms = st_geometry(nodes)
+    # For each replacement edge:
+    # --> Merge geometries of the edges it replaces into a single geometry.
     merge_geoms = function(E) {
       orig_edges = E$.tidygraph_edge_index
       orig_geoms = edge_geoms[orig_edges]
       new_geom = st_line_merge(st_combine(orig_geoms))
-      # There is one situation where merging lines like this is problematic.
-      # That is when the source and target node of the new edge are the same.
-      # Hence, the original edges to be merged form a closed loop.
-      # Any original edge endpoint can then be the startpoint of the new edge.
-      # st_line_merge chooses the point with the lowest x coordinate.
-      # This is not necessarily the source node we defined.
-      # This behaviour comes from third partly libs and can not be tuned.
-      # Hence, we manually need to reorder the points in the merged line.
-      src = E$from
-      trg = E$to
-      if (src == trg && length(orig_edges) > 1) {
+      # There are two situations where merging lines like this is problematic.
+      # 1. When the source and sink node of the new edge are the same.
+      # --> In this case the original edges to be replaced form a closed loop.
+      # --> Any original endpoint can then be the startpoint of the new edge.
+      # --> st_line_merge chooses the point with the lowest x coordinate.
+      # --> This is not necessarily the source node we defined.
+      # --> This behaviour comes from third partly libs and can not be tuned.
+      # --> Hence, we manually need to reorder the points in the merged line.
+      if (E$from == E$to && length(orig_edges) > 1) {
         pts = st_cast(new_geom, "POINT")
-        src_idx = st_equals(node_geoms[src], pts)[[1]]
-        if (length(src_idx) == 1) {
+        from_idx = st_equals(node_geoms[E$from], pts)[[1]]
+        if (length(from_idx) == 1) {
           n = length(pts)
-          ordered_pts = c(pts[c(src_idx:n)], pts[c(2:src_idx)])
+          ordered_pts = c(pts[c(from_idx:n)], pts[c(2:from_idx)])
           new_geom = st_cast(st_combine(ordered_pts), "LINESTRING")
         }
       }
+      # 2. When the new edge crosses itself.
+      # --> In this case st_line_merge creates a multilinestring geometry.
+      # --> We just want a regular linestring (even if this is invalid).
+      if (any(st_is(new_geom, "MULTILINESTRING"))) {
+        new_geom = multilinestrings_to_linestrings(new_geom)
+      }
       new_geom
     }
-    new_geoms = do.call("c", lapply(new_edge_list, merge_geoms))
-    # Add the geometries to the new edges data frame.
-    # Use the same geometry column name as in the original edges data frame.
-    geom_colname = attr(edges, "sf_column")
-    new_edges[geom_colname] = list(new_geoms)
-    new_edges = st_as_sf(new_edges, sf_column_name = geom_colname)
+    new_geoms = do.call("c", lapply(new_idxs, merge_geoms))
   }
-  ## ========================================
-  # STEP IV: ADD MERGED EDGES TO THE NETWORK
+  ## ============================================
+  # STEP V: ADD REPLACEMENT EDGES TO THE NETWORK
   # The newly created edges should be added to the original network.
   # This must happen before removing the pseudo nodes.
-  # Otherwise, the source and target indices do not match their nodes anymore.
-  ## ========================================
-  # Bind the original and new edges.
-  edges$.tidygraph_edge_index = as.list(edges$.tidygraph_edge_index)
-  all_edges = bind_rows(edges, new_edges)
+  # Otherwise their from and to values do not match the correct node indices.
+  ## ============================================
+  # Create the data frame for the new edges.
+  new_edges = cbind(
+    data.frame(do.call("rbind", new_idxs)),
+    data.frame(do.call("rbind", new_attrs))
+  )
+  new_edges[geom_colname] = list(new_geoms)
+  # Bind together with the original edges.
+  # Merged edges may have list-columns for some attributes.
+  # This requires a bit more complicated rowbinding.
+  all_edges = bind_rows_list(edges, new_edges)
+  if (spatial) all_edges = st_as_sf(all_edges, sf_column_name = geom_colname)
   # Recreate an sfnetwork.
   x_new = sfnetwork_(nodes, all_edges, directed = directed)
   ## ============================================
-  # STEP V: REMOVE PSEUDO NODES FROM THE NETWORK
+  # STEP VI: REMOVE PSEUDO NODES FROM THE NETWORK
   # Remove all the detected pseudo nodes from the original network.
   # This will automatically also remove their incident edges.
   # Remember that their replacement edges have already been added in step IV.
+  # From and to indices will be updated automatically.
   ## ============================================
   x_new = delete_vertices(x_new, pseudo) %preserve_all_attrs% x
-  ## =============================================
-  # STEP VI: STORE ORIGINAL EDGE DATA IF REQUESTED
+  ## ==============================================
+  # STEP VII: STORE ORIGINAL EDGE DATA IF REQUESTED
   # Users can request to store the data of original edges in a special column.
   # This column will - by tidygraph design - be named .orig_data.
   # The value in this column is for each edge a tibble containing:
   # --> The data of the original edges that were merged into the new edge.
-  ## =============================================
+  ## ==============================================
   if (store_original_data) {
     # Store the original edge data in a .orig_data column.
     new_edges = edges_as_sf(x_new)
     edges$.tidygraph_edge_index = NULL
     copy_data = function(i) edges[i, , drop = FALSE]
     new_edges$.orig_data = lapply(new_edges$.tidygraph_edge_index, copy_data)
-    edge_graph_attributes(x_new) = new_edges
+    edge_attribute_values(x_new) = new_edges
   }
   # Return in a list.
   list(
@@ -761,13 +952,15 @@ to_spatial_smooth = function(x, store_original_data = FALSE) {
 #' edges are connected at the points of subdivision. Returns a
 #' \code{morphed_sfnetwork} containing a single element of class
 #' \code{\link{sfnetwork}}. This morpher requires edges to be spatially
-#' explicit.
+#' explicit and nodes to be spatially unique (i.e. not more than one node at
+#' the same spatial location).
 #' @importFrom igraph is_directed
-#' @importFrom sf st_crs st_geometry
+#' @importFrom sf st_crs st_crs<- st_geometry st_geometry<- st_precision
+#' st_precision<-
 #' @importFrom sfheaders sf_to_df sfc_linestring sfc_point
 #' @export
 to_spatial_subdivision = function(x) {
-  require_spatially_explicit_edges(x)
+  require_explicit_edges(x)
   if (will_assume_constant(x)) raise_assume_constant("to_spatial_subdivision")
   # Retrieve nodes and edges from the network.
   nodes = nodes_as_sf(x)
@@ -843,6 +1036,7 @@ to_spatial_subdivision = function(x) {
   # Build the new edge geometries.
   new_edge_geoms = sfc_linestring(new_edge_coords, linestring_id = "edge_id")
   st_crs(new_edge_geoms) = st_crs(edges)
+  st_precision(new_edge_geoms) = st_precision(edges)
   new_edge_coords$edge_id = NULL
   ## ===================================
   # STEP V: CONSTRUCT THE NEW EDGE DATA
@@ -874,6 +1068,7 @@ to_spatial_subdivision = function(x) {
   is_new_boundary = rep(is_split | is_boundary, reps)
   new_node_geoms = sfc_point(new_edge_coords[is_new_boundary, ])
   st_crs(new_node_geoms) = st_crs(nodes)
+  st_precision(new_node_geoms) = st_precision(nodes)
   ## =====================================
   # STEP VII: CONSTRUCT THE NEW NODE DATA
   # We now have the geometries of the new nodes.
@@ -927,7 +1122,7 @@ to_spatial_subdivision = function(x) {
   x_new = sfnetwork_(new_nodes, new_edges, directed = directed)
   # Return in a list.
   list(
-    subdivision = x_new %preserve_graph_attrs% x
+    subdivision = x_new %preserve_network_attrs% x
   )
 }
 
